/***************************************************************************
 *   Copyright (C) 2007 by Chunkeey                                        *
 *   chunkeey@web.de                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <usb.h>

#include "hw.h"
#include "asusled.h"
#include "SDL.h"
#include "SDL_image.h"
#define VERSION "0.3"

static int convert_image(unsigned char *dest, const unsigned char *src, int x_offset, int y_offset, int width)
{
  int x,y,z = 0;
  unsigned char buf;

  for (y = 0; y < 16; y+=8)
    for (x = 0; x < 128; ++x) {
      for (z = buf = 0; z < 8; z++) {
        buf |= ((src[(y+z+y_offset) * width + x + x_offset]) != 0) << z;
      }
      dest[(127-x) + ((y >> 3) << 7)] = buf;
    }

  return EXIT_SUCCESS;
}

#define min(a,b) (a > b) ? (b) : (a)

static unsigned char* recalc_image(SDL_Surface* image, int cutoff)
{
  int x, y, z, a;
  unsigned char *tmp, *src;
  unsigned int col = 0;

  tmp = malloc(image->w * image->h);
  if (!tmp) {
    fprintf(stderr, "Out of memory!\n");
    return NULL;
  }

  src = (unsigned char*)(image->pixels);
  a = min(image->format->BytesPerPixel, 3);
  for (y = 0; y < image->h; y++)
   for (x = 0; x < image->w; x++) {
    for (z = 0; z < a; z++)
        col |= src[((y * (image->w)) + x) * image->format->BytesPerPixel + z];

    tmp[(y * image->w) + x] = (col > cutoff);
    col = 0;
  }

  return tmp;
}

struct usb_dev_handle *oled_find_device()
{
  struct usb_dev_handle* handle;
  struct usb_device* usb_dev = NULL;
  struct usb_bus *busses;
  struct usb_bus *bus;

  usb_init();
  usb_find_busses();
  usb_find_devices();

  busses = usb_get_busses();

  if (!busses)
    return NULL;

  for (bus = busses; bus; bus = bus->next) {
    struct usb_device *dev;

    for (dev = bus->devices; dev; dev = dev->next) {
      if ((dev->descriptor.idVendor == 0x0b05) && (dev->descriptor.idProduct == 0x1726)) {
        usb_dev = dev;
      }
    }
  }

  if (!usb_dev) {
    fprintf(stderr, "Could not find any ASUS LCM device!\n");
    return NULL;
  }

  if (!(handle = usb_open(usb_dev))) {
    fprintf(stderr, "Could not open USB device\n");
    return NULL;
  }
	
  if (usb_claim_interface(handle, 0) < 0) {
    fprintf(stderr, "Could not claim USB interface\n");
    fprintf(stderr, "aren't you root? \"rmmod usbhid\" and try again...\n");
    return NULL;
  }

  return handle;
}

int oled_close(struct usb_dev_handle *handle)
{
  usb_release_interface(handle, 0);
  usb_close(handle);

  return EXIT_SUCCESS;
}

int oled_control(struct usb_dev_handle *handle, int on_off)
{
  struct oled_packet buffer = { { 0 }, { 0 } };
  int ret = -1;

  switch (on_off) {
    case DISPLAY_OFF:
      FILL_HEADER(buffer.header, 0x20, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00);
      buffer.bitmap[0] = 0xae;
      break;
    case DISPLAY_ON:
      FILL_HEADER(buffer.header, 0x20, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00);
      buffer.bitmap[0] = 0xaf;
      break;
    default:
      return -EINVAL; 
      break; 
  }

  ret = usb_bulk_write(handle, 2, (char*)&buffer, sizeof(buffer), -1);
  if (ret < 0) {
    fprintf(stderr, "Could not alter display state.\n");
    fprintf(stderr, "If it doesn't respond at all anymore"
                    " just do a full powercycle!");
  }
  return ret;
}

int oled_draw_image(struct usb_dev_handle *handle, int func,
		    int width, int height, const unsigned char *bitmap)
{
  struct oled_packet *buf = calloc(width / 64, sizeof(*buf));
  int ret = -1, i;

  if (bitmap) {
    for (i = 0; i < (width / 128 * height / 16) ; i++) {
      switch (func) {
      case DRAW_ROLL: 
          FILL_HEADER(buf[i].header, 0x40, 0x80, width / 64, i + 1, 0x00, 0x01, 0xff);
          break;
      case DRAW_STATIC:
          FILL_HEADER(buf[i].header, 0x10 + i , 0x80, 0x01, 0x01, 0x00, 0x01, 0x00);
          break;
      case DRAW_FLASH:
          FILL_HEADER(buf[i].header, 0x10, 0x80, 0x01, 0x01, 0x0, 0x01, 0xff);	
          break;
      }
      convert_image(buf[i].bitmap, bitmap, (i / 2) * 128, (i % 2) * 16, width);
    }
  }
  else {
    memset(buf[0].bitmap, 0xff, sizeof(buf[0].bitmap));
    memset(buf[1].bitmap, 0xff, sizeof(buf[1].bitmap));
    FILL_HEADER(buf[0].header, 0x10, 0x80, 0x01, 0x01, 0x00, 0x01, 0x00);
    FILL_HEADER(buf[1].header, 0x11, 0x80, 0x01, 0x01, 0x00, 0x01, 0x00);
  }

  for (i = 0; i < (width / 64);i++) {
    while (ret < 0) {
        ret = usb_interrupt_write(handle, 2, (char*)&buf[i], sizeof(buf[i]), -1);
    }
    ret = -1;
  }

  ret = 0;

  free(buf);

  return ret;
}

static int check_image_metadata(int width, int height, int func)
{
  if ((func == DRAW_FLASH) && ((height != 16) || (width != 128))) {
    fprintf(stderr, "Picture is not 128x16!\n");
    return -1;
  } else if ((func == DRAW_STATIC) && ((height != 32) || (width != 128))) {
    fprintf(stderr, "Picture is not 128x32!\n"); 
    return -1;
  } else if ((func == DRAW_ROLL) && ((height != 32) || (width < 128) || (width % 128) || (width > 1792))) {
    fprintf(stderr, "Picture is not at least 128x32, bigger than 1792x32, width is not a multiple of 128 or height is not 32!\n");
    return -1;
  } else
   return 0;
}

static int process_image(struct usb_dev_handle *handle, const char *filename, int func)
{
  SDL_Surface* image;
  unsigned char* bitmp;

  image = IMG_Load(filename);
  if (!image) {
    fprintf(stderr, "%s\n", IMG_GetError());
    return errno;
  }

  if (check_image_metadata(image->w, image->h, func)) {
    fprintf(stderr, "Image dimension mismatch!\n");
    SDL_FreeSurface(image);
    return -EINVAL;
  }

  bitmp = recalc_image(image, 128);
  SDL_FreeSurface(image);
  if (bitmp) {
    oled_draw_image(handle, func, image->w, image->h, bitmp);
    free(bitmp);

    if ((func == DRAW_ROLL) || (func == DRAW_FLASH)) {
      fprintf(stdout, "Press [Enter] to continue\n");
      getchar();
    }
  } else {
    return -1;
  }

  return EXIT_SUCCESS;
}

static int show_version(void)
{
  fprintf(stdout, "ASUS' USB OLED Display Driver Version: %s\n", VERSION);
  return EXIT_SUCCESS;
}

static int show_options(char *file)
{
  fprintf(stdout, "Usage: %s [ -ed ] [ -srf picture ]\n", file);
  fprintf(stdout, "-e\t\tenable display\n");
  fprintf(stdout, "-d\t\tdisable display\n");
  fprintf(stdout, "-s\t\tdisplay \"picture\"\n");
  fprintf(stdout, "-r\t\tdisplay \"picture\" with a special effect ;)\n");
  fprintf(stdout, "-f\t\tdisplay \"picture\" flashing\n");  
  fprintf(stdout, "-V\t\tshow version\n");
  fprintf(stdout, "-h\t\tshow this usage message\n");
  return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
  struct usb_dev_handle *handle;
  int ret;

  if ((argc < 2) || (argc > 3) || (!strncmp(argv[1], "-h", 2)))
    return show_options(argv[0]);

 if (!strncmp(argv[1], "-V", 2))
     return show_version();

  if (!(handle = oled_find_device()))
    return -1;

  if (!strncmp(argv[1], "-e", 2))
    ret = oled_control(handle, DISPLAY_OFF) | oled_control(handle, DISPLAY_ON);
  else if (!strncmp(argv[1], "-d", 2))
    ret = oled_control(handle, DISPLAY_ON) | oled_control(handle, DISPLAY_OFF);
  else if (!strncmp(argv[1], "-r", 2) && (argc == 3))
    ret = process_image(handle, argv[2], DRAW_ROLL);
  else if (!strncmp(argv[1], "-s", 2) && (argc == 3))
    ret = process_image(handle, argv[2], DRAW_STATIC);
  else if (!strncmp(argv[1], "-f", 2) && (argc == 3))
    ret = process_image(handle, argv[2], DRAW_FLASH);
  else {
    fprintf(stderr, "Invalid Parameters!\n");
    show_options(argv[0]);
  }

  oled_close(handle);

  return EXIT_SUCCESS;
}
